from datetime import datetime

class OrderedList:
    def __init__(self,list_of_values) -> None:
        self._data = self._basic_method(list_of_values=list_of_values)
    
    def __repr__(self) -> str:
        return self.tail()
    
    def _basic_method(self,list_of_values):
        mod_func = lambda date_string: datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S%z').replace(tzinfo=None)
        nd = {'datetime':[],'open':[],'high':[],'low':[],'close':[],'volume':[],'oi':[]}
        for row in list_of_values:
            nd['datetime'].append(mod_func(row[0]))
            nd['open'].append(row[1])
            nd['high'].append(row[2])
            nd['low'].append(row[3])
            nd['close'].append(row[4])
            nd['volume'].append(row[5])
            nd['oi'].append(row[6])
        return nd
    
    def print_dict_as_table(self,data):
        max_lengths = [max(len(str(value)) for value in col) for col in data.values()]
        header_row = "|".join(header.center(length) for header, length in zip(data.keys(), max_lengths))
        print(header_row)
        separator = "+".join("-" * length for length in max_lengths)
        print(separator)
        for i in range(len(next(iter(data.values())))):
            row_values = [str(data[key][i]).center(length) for key, length in zip(data.keys(), max_lengths)]
            row = "|".join(row_values)
            print(row)

    def head(self, val=5):
        data_head = {key: value[:val] for key, value in self._data.items()}
        self.print_dict_as_table(data_head)
    
    def iloc(self,start, end):
        data_head = {key: value[start:end] for key, value in self._data.items()}
        self.print_dict_as_table(data_head)
    
    def tail(self, val=5):
        data_head = {key: value[-val:] for key, value in self._data.items()}
        self.print_dict_as_table(data_head)


import urllib.parse
import datetime
from zern.utils.Types import MODE_VALUES,SEGMENT

def encodeURIComponent(string):
    o = {"!": "%21", "'": "%27", "(": "%28", ")": "%29", "~": "%7E", "%20": "+", "%00": "\0"}
    encoded_string = urllib.parse.quote(string,safe='*')
    for char, encoded_char in o.items():
        encoded_string = encoded_string.replace(char, encoded_char)
    return encoded_string


def buf2Uint32(data):
    if len(data) != 4:
        raise ValueError("Input data must have exactly 4 bytes")
    uint32_value = int.from_bytes(data, byteorder='big')
    return uint32_value

def buf2Uint16(data):
    if len(data) != 2:
        raise ValueError("Input data must have exactly 2 bytes")
    uint16_value = int.from_bytes(data, byteorder='big')
    return uint16_value

def split_packets(data):
    t = 2
    packets = []
    length = buf2Uint16(data[:2])
    for _ in range(length):
        packet_length = buf2Uint16(data[t:t + 2])
        packet_data = data[t + 2:t + 2 + packet_length]
        packets.append(packet_data)
        t += 2 + packet_length
    return packets
import datetime

def date_to_string(date_obj):
    year = str(date_obj.year)
    month = str(date_obj.month).zfill(2)
    day = str(date_obj.day).zfill(2)
    hour = str(date_obj.hour).zfill(2)
    minute = str(date_obj.minute).zfill(2)
    second = str(date_obj.second).zfill(2)
    return f"{year}-{month}-{day} {hour}:{minute}:{second}"

def calculate_change(entry):
    t = s = i = n = 0
    if 'closePrice' in entry.keys():
        s = entry['lastPrice'] - entry['closePrice']
        t = 100 * s / entry['closePrice']
    if 'openPrice' in entry.keys():
        i = entry['lastPrice'] - entry['openPrice']
        n = 100 * i / entry['openPrice']
    return {
        'change': t,
        'absoluteChange': s,
        'openChange': i,
        'openChangePercent': n
    }


def parse_binary(data):
    if len(data) == 1:
        return None
    packets = split_packets(data)
    result = []
    for packet in packets:
        token = buf2Uint32(packet[:4])
        segment = token & 255
        divisor = 100
        
        
        if segment == SEGMENT.segmentNseCD:
            divisor = 1e7
        elif not (segment != SEGMENT.segmentBseCD and segment != SEGMENT.segmentNseCOM):
            divisor = 1e4
        if len(packet) == 8:
            result.append({
                'mode': MODE_VALUES.modeLTP,
                'isTradeable': True,
                'token': token,
                'lastPrice': buf2Uint32(packet[4:8]) / divisor
            })
        elif len(packet) == 12:
            entry = {
                'mode': MODE_VALUES.modeLTPC,
                'isTradeable': True,
                'token': token,
                'lastPrice': buf2Uint32(packet[4:8]) / divisor,
                'closePrice': buf2Uint32(packet[8:12]) / divisor
            }
            entry.update(calculate_change(entry))
            result.append(entry)
        elif len(packet) in [28, 32]:
            entry = {
                'mode': MODE_VALUES.modeFull,
                'isTradeable': False,
                'token': token,
                'lastPrice': buf2Uint32(packet[4:8]) / divisor,
                'highPrice': buf2Uint32(packet[8:12]) / divisor,
                'lowPrice': buf2Uint32(packet[12:16]) / divisor,
                'openPrice': buf2Uint32(packet[16:20]) / divisor,
                'closePrice': buf2Uint32(packet[20:24]) / divisor
            }
            entry.update(calculate_change(entry))
            result.append(entry)
        elif len(packet) == 492:
            entry = {
                'mode': MODE_VALUES.modeFull,
                'token': token,
                'extendedDepth': {
                    'buy': [],
                    'sell': []
                }
            }
            offset = 12
            depth_data = packet[offset:492]
            for _ in range(40):
                quantity = buf2Uint32(depth_data[:4])
                price = buf2Uint32(depth_data[4:8]) / divisor
                orders = buf2Uint32(depth_data[8:12])
                entry['extendedDepth']['buy' if _ < 20 else 'sell'].append({
                    'quantity': quantity,
                    'price': price,
                    'orders': orders
                })
                depth_data = depth_data[12:]
            result.append(entry)
        elif len(packet) in [164, 184]:
            entry = {
                'mode': MODE_VALUES.modeQuote,
                'token': token,
                'isTradeable': True,
                'volume': buf2Uint32(packet[16:20]),
                'lastQuantity': buf2Uint32(packet[8:12]),
                'totalBuyQuantity': buf2Uint32(packet[20:24]),
                'totalSellQuantity': buf2Uint32(packet[24:28]),
                'lastPrice': buf2Uint32(packet[4:8]) / divisor,
                'averagePrice': buf2Uint32(packet[12:16]) / divisor,
                'openPrice': buf2Uint32(packet[28:32]) / divisor,
                'highPrice': buf2Uint32(packet[32:36]) / divisor,
                'lowPrice': buf2Uint32(packet[36:40]) / divisor,
                'closePrice': buf2Uint32(packet[40:44]) / divisor
            }
            entry.update(calculate_change(entry))
            if len(packet) == 184:
                offset = 44
                entry['lastTradedTime'] = date_to_string(datetime.datetime.fromtimestamp(buf2Uint32(packet[44:48])))
                entry['oi'] = buf2Uint32(packet[48:52])
                entry['oiDayHigh'] = buf2Uint32(packet[52:56])
                entry['oiDayLow'] = buf2Uint32(packet[56:60])
            else:
                offset = 64
            depth_data = packet[offset:offset+120]
            entry['mode'] = MODE_VALUES.modeFull
            entry['depth'] = {'buy': [], 'sell': []}
            for _ in range(10):
                quantity = buf2Uint32(depth_data[:4])
                price = buf2Uint32(depth_data[4:8]) / divisor
                orders = buf2Uint16(depth_data[8:10])
                entry['depth']['buy' if _ < 5 else 'sell'].append({
                    'quantity': quantity,
                    'price': price,
                    'orders': orders
                })
                depth_data = depth_data[12:]
            result.append(entry)
    return result



import requests
from time import sleep
import random
import pyotp

class Session:
    TYPE_GET = requests.get
    TYPE_POST = requests.post

    def __init__(self,user_name, password, totp_key) -> None:
        self.user_name = user_name
        self._private__password = password
        self.totp_key = totp_key
        self.cookies = None
        self.headers = {
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8',
            'accept-language': 'en-GB,en;q=0.8',
            'cache-control': 'max-age=0',
            'if-modified-since': 'Thu, 17 Oct 2019 07:18:26 GMT',
            'if-none-match': '"3147526947+gzip"',
            'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36',
        }
        self._base_url = 'https://kite.zerodha.com'
        self._login_url = 'https://kite.zerodha.com/api/login'
        self._totp_url = 'https://kite.zerodha.com/api/twofa'
        self._enc_token = None
        self.start()
        self._get_instruments()

    
    def start(self):
        resp = requests.get(self._base_url,headers=self.headers)
        self.cookies = resp.cookies
        sleep(random.random())
        print('login: waiting for almost 5 seconds')
        login_data = {'user_id': self.user_name,'password': self._private__password}
        resp = requests.post(self._login_url, cookies=self.cookies, headers=self.headers, data=login_data)
        request_id = resp.json()['data']['request_id']
        sleep(random.random())
        print('auto TOTP: waiting for almost 5 seconds')
        totp_data = {'user_id': self.user_name,'request_id': request_id,'twofa_value': pyotp.TOTP(self.totp_key).now(),'twofa_type': 'totp','skip_session': ''}
        response = requests.post('https://kite.zerodha.com/api/twofa', cookies=self.cookies, headers=self.headers, data=totp_data)
        self.cookies = response.cookies
        self._enc_token = self.cookies.get("enctoken")
        self.headers['authorization'] = 'enctoken {}'.format(self._enc_token)

        sleep(random.random())
        done = True
        try:
            success = response.json()['status']
        except:
            done = False
        if done:
            print('Authentication: {}'.format(success))
        else:
            print('Authentication: failed')
    
    def request(self,url,method=TYPE_GET,cookies=True,headers=True,data=None,params=None):
        response = method(url,
                          cookies=self.cookies if cookies else None,
                          headers=self.headers if headers else None,
                          data = data,
                          params=params)
        return response.json()

    def _get_instruments(self):
        new_headers = {
            'accept': 'application/json, text/plain, */*',
            'accept-language': 'en-GB,en;q=0.9',
            'kite-iframe': 'kite_iframe_user',
            'origin': 'https://insights.sensibull.com',
            'referer': 'https://insights.sensibull.com/',
            'sec-ch-ua': '"Brave";v="123", "Not:A-Brand";v="8", "Chromium";v="123"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Linux"',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-site',
            'sec-gpc': '1',
            'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36',
        }
        response = requests.get('https://oxide.sensibull.com/v1/compute/cache/insights/instrument_metacache/1', headers=new_headers)
        print('getting instrument tokens')
        self.instruments = response.json()
        #sleep(2)
        